/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.commons.csv;

import static org.apache.commons.csv.Constants.COMMA;
import static org.apache.commons.csv.Constants.CR;
import static org.apache.commons.csv.Constants.CRLF;
import static org.apache.commons.csv.Constants.DOUBLE_QUOTE;
import static org.apache.commons.csv.Constants.ESCAPE;
import static org.apache.commons.csv.Constants.LF;
import static org.apache.commons.csv.Constants.TAB;

import java.io.IOException;
import java.io.Reader;
import java.io.Serializable;
import java.io.StringWriter;

import org.apache.commons.csv.CSVFormat.CSVFormatBuilder;

/**
 * The format specification of a CSV file.
 *
 * This class is immutable.
 * 
 * @version $Id: $
 */
public class CSVFormat implements Serializable {

    private static final long serialVersionUID = 1L;

    private final char delimiter;
    private final Character quoteChar;
    private final Quote quotePolicy;
    private final Character commentStart;
    private final Character escape;
    private final boolean ignoreSurroundingSpaces; // Should leading/trailing spaces be ignored around values?
    private final boolean ignoreEmptyLines;
    private final String recordSeparator; // for outputs
    private final String[] header;

    /**
     * Comma separated format as defined by <a href="http://tools.ietf.org/html/rfc4180">RFC 4180</a>.
     * <ul>
     * <li>withDelimiter(',')</li>
     * <li>withQuoteChar('"')</li>
     * <li>withLineSeparator(CRLF)</li>
     * </ul>
     */
    public static final CSVFormat RFC4180 =
            defaults()
            .withIgnoreEmptyLines(false)
            .build();

    /**
     * Standard comma separated format, as for {@link #RFC4180} but allowing blank lines.
     * <ul>
     * <li>withDelimiter(',')</li>
     * <li>withQuoteChar('"')</li>
     * <li>withEmptyLinesIgnored(true)</li>
     * <li>withLineSeparator(CRLF)</li>
     * </ul>
     */
    public static final CSVFormat DEFAULT = // TODO rename to something more meaningful
            defaults()
            .build();

    /**
     * Excel file format (using a comma as the value delimiter). Note that the actual value delimiter used by Excel is
     * locale dependent, it might be necessary to customize this format to accomodate to your regional settings.
     * <p/>
     * For example for parsing or generating a CSV file on a French system the following format will be used:
     *
     * <pre>
     * CSVFormat fmt = CSVFormat.EXCEL.withDelimiter(';');
     * </pre>
     * Settings are:
     * <ul>
     * <li>withDelimiter(',')</li>
     * <li>withQuoteChar('"')</li>
     * <li>withLineSeparator(CRLF)</li>
     * </ul>
     * Note: this is currently the same as RFC4180
     */
    public static final CSVFormat EXCEL =
            defaults()
            .withIgnoreEmptyLines(false)
            .build();

    /** Tab-delimited format, with quote; leading and trailing spaces ignored. */
    public static final CSVFormat TDF =
            defaults()
            .withDelimiter(TAB)
            .withIgnoreSurroundingSpaces(true)
            .build();

    /**
     * Default MySQL format used by the <tt>SELECT INTO OUTFILE</tt> and <tt>LOAD DATA INFILE</tt> operations. This is
     * a tab-delimited format with a LF character as the line separator. Values are not quoted and special characters
     * are escaped with '\'.
     *
     * @see <a href="http://dev.mysql.com/doc/refman/5.1/en/load-data.html">
     *      http://dev.mysql.com/doc/refman/5.1/en/load-data.html</a>
     */
    public static final CSVFormat MYSQL =
            defaults()
            .withDelimiter(TAB)
            .withQuoteChar(null)
            .withEscape(ESCAPE)
            .withIgnoreEmptyLines(false)
            .withRecordSeparator(LF)
            .build();

    /**
     * Factory method for creating CSV formats.
     * 
     * @param delimiter 
     *            the char used for value separation, must not be a line break character
     * @throws IllegalArgumentException if the delimiter is a line break character
     */
    public static CSVFormatBuilder newFormat(char delimiter) {
        return new CSVFormatBuilder(delimiter);
    }

    /**
     * Standard comma separated format, as for {@link #RFC4180} but allowing blank lines.
     * <ul>
     * <li>withDelimiter(',')</li>
     * <li>withQuoteChar('"')</li>
     * <li>withEmptyLinesIgnored(true)</li>
     * <li>withLineSeparator(CRLF)</li>
     * </ul>
     */
    public static CSVFormatBuilder defaults() {
        return new CSVFormatBuilder(COMMA, DOUBLE_QUOTE, null, null, null, false, true, CRLF, null);
    }

    /**
     * Creates a customized CSV format.
     *
     * @param delimiter
     *            the char used for value separation, must not be a line break character
     * @param quoteChar
     *            the char used as value encapsulation marker
     * @param quotePolicy 
     *            the quote policy
     * @param commentStart
     *            the char used for comment identification
     * @param escape
     *            the char used to escape special characters in values
     * @param ignoreSurroundingSpaces
     *            <tt>true</tt> when whitespaces enclosing values should be ignored
     * @param ignoreEmptyLines
     *            <tt>true</tt> when the parser should skip empty lines
     * @param recordSeparator
     *            the line separator to use for output
     * @param header
     *            the header
     * @throws IllegalArgumentException if the delimiter is a line break character
     */
    private CSVFormat(final char delimiter, final Character quoteChar, final Quote quotePolicy, final Character commentStart, final Character escape, final 
                    boolean ignoreSurroundingSpaces, final boolean ignoreEmptyLines, final String lineSeparator, 
            final String[] header) {
        if (isLineBreak(delimiter)) {
            throw new IllegalArgumentException("The delimiter cannot be a line break");
        }
        this.delimiter = delimiter;
        this.quoteChar = quoteChar;
        this.quotePolicy = quotePolicy;
        this.commentStart = commentStart;
        this.escape = escape;
        this.ignoreSurroundingSpaces = ignoreSurroundingSpaces;
        this.ignoreEmptyLines = ignoreEmptyLines;
        this.recordSeparator = lineSeparator;
        this.header = header;
    }

    /**
     * Returns true if the given character is a line break character.
     *
     * @param c
     *            the character to check
     *
     * @return true if <code>c</code> is a line break character
     */
    private static boolean isLineBreak(final Character c) {
        return c != null && isLineBreak(c.charValue());
    }

    /**
     * Returns true if the given character is a line break character.
     *
     * @param c
     *            the character to check
     *
     * @return true if <code>c</code> is a line break character
     */
    private static boolean isLineBreak(final char c) {
        return c == LF || c == CR;
    }

    /**
     * Returns the character delimiting the values (typically ';', ',' or '\t').
     *
     * @return the delimiter character
     */
    public char getDelimiter() {
        return delimiter;
    }

    /**
     * Returns the character used to encapsulate values containing special characters.
     *
     * @return the quoteChar character
     */
    public Character getQuoteChar() {
        return quoteChar;
    }

    /**
     * Returns whether an quoteChar has been defined.
     *
     * @return {@code true} if an quoteChar is defined
     */
    public boolean isQuoting() {
        return quoteChar != null;
    }

    /**
     * Returns the character marking the start of a line comment.
     *
     * @return the comment start marker.
     */
    public Character getCommentStart() {
        return commentStart;
    }

    /**
     * Specifies whether comments are supported by this format.
     *
     * Note that the comment introducer character is only recognised at the start of a line.
     *
     * @return <tt>true</tt> is comments are supported, <tt>false</tt> otherwise
     */
    public boolean isCommentingEnabled() {
        return commentStart != null;
    }

    /**
     * Returns the escape character.
     *
     * @return the escape character
     */
    public Character getEscape() {
        return escape;
    }

    /**
     * Returns whether escape are being processed.
     *
     * @return {@code true} if escapes are processed
     */
    public boolean isEscaping() {
        return escape != null;
    }

    /**
     * Specifies whether spaces around values are ignored when parsing input.
     *
     * @return <tt>true</tt> if spaces around values are ignored, <tt>false</tt> if they are treated as part of the
     *         value.
     */
    public boolean getIgnoreSurroundingSpaces() {
        return ignoreSurroundingSpaces;
    }

    /**
     * Specifies whether empty lines between records are ignored when parsing input.
     *
     * @return <tt>true</tt> if empty lines between records are ignored, <tt>false</tt> if they are turned into empty
     *         records.
     */
    public boolean getIgnoreEmptyLines() {
        return ignoreEmptyLines;
    }

    /**
     * Returns the line separator delimiting output records.
     *
     * @return the line separator
     */
    public String getRecordSeparator() {
        return recordSeparator;
    }

    String[] getHeader() {
        return header;
    }

    /**
     * Parses the specified content.
     *
     * @param in
     *            the input stream
     */
    public Iterable<CSVRecord> parse(final Reader in) throws IOException {
        return new CSVParser(in, this);
    }

    /**
     * Format the specified values.
     *
     * @param values
     *            the values to format
     */
    public String format(final Object... values) {
        final StringWriter out = new StringWriter();
        try {
            new CSVPrinter(out, this).printRecord(values);
            return out.toString().trim();
        } catch (final IOException e) {
            // should not happen because a StringWriter does not do IO.
            throw new IllegalStateException(e);
        }
    }

    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        sb.append("Delimiter=<").append(delimiter).append('>');
        if (isEscaping()) {
            sb.append(' ');
            sb.append("Escape=<").append(escape).append('>');
        }
        if (isQuoting()) {
            sb.append(' ');
            sb.append("QuoteChar=<").append(quoteChar).append('>');
        }
        if (isCommentingEnabled()) {
            sb.append(' ');
            sb.append("CommentStart=<").append(commentStart).append('>');
        }
        if (getIgnoreEmptyLines()) {
            sb.append(" EmptyLines:ignored");
        }
        if (getIgnoreSurroundingSpaces()) {
            sb.append(" SurroundingSpaces:ignored");
        }
        return sb.toString();
    }

    /**
     * Returns the quote policy output fields.
     *
     * @return the quote policy
     */
    public Quote getQuotePolicy() {
        return quotePolicy;
    }
    
    public static class CSVFormatBuilder {

        private char delimiter;
        private Character quoteChar;
        private Quote quotePolicy;
        private Character commentStart;
        private Character escape;
        private boolean ignoreSurroundingSpaces; // Should leading/trailing spaces be ignored around values?
        private boolean ignoreEmptyLines;
        private String recordSeparator; // for outputs
        private String[] header;

        /**
         * Creates a customized CSV format.
         *
         * @param delimiter
         *            the char used for value separation, must not be a line break character
         * @param quoteChar
         *            the char used as value encapsulation marker
         * @param quotePolicy 
         *            the quote policy
         * @param commentStart
         *            the char used for comment identification
         * @param escape
         *            the char used to escape special characters in values
         * @param ignoreSurroundingSpaces
         *            <tt>true</tt> when whitespaces enclosing values should be ignored
         * @param ignoreEmptyLines
         *            <tt>true</tt> when the parser should skip empty lines
         * @param recordSeparator
         *            the line separator to use for output
         * @param header
         *            the header
         * @throws IllegalArgumentException if the delimiter is a line break character
         */
        // package protected for use by test code
        CSVFormatBuilder(final char delimiter, final Character quoteChar, final Quote quotePolicy, final Character commentStart, final Character escape, final 
                        boolean ignoreSurroundingSpaces, final boolean ignoreEmptyLines, final String lineSeparator, 
                final String[] header) {
            if (isLineBreak(delimiter)) {
                throw new IllegalArgumentException("The delimiter cannot be a line break");
            }
            this.delimiter = delimiter;
            this.quoteChar = quoteChar;
            this.quotePolicy = quotePolicy;
            this.commentStart = commentStart;
            this.escape = escape;
            this.ignoreSurroundingSpaces = ignoreSurroundingSpaces;
            this.ignoreEmptyLines = ignoreEmptyLines;
            this.recordSeparator = lineSeparator;
            this.header = header;
        }

        /**
         * Creates a basic CSVFormatBuilder.
         * 
         * @param delimiter 
         *            the char used for value separation, must not be a line break character
         * @throws IllegalArgumentException if the delimiter is a line break character
         */
        private CSVFormatBuilder(char delimiter){
            this(delimiter, null, null, null, null, false, false, null, null);
        }

        public CSVFormat build() {
            validate();
            return new CSVFormat(delimiter, quoteChar, quotePolicy, commentStart, escape,
                                 ignoreSurroundingSpaces, ignoreEmptyLines, recordSeparator, header);
        }
        
        /**
         * Verifies the consistency of the parameters and throws an IllegalStateException if necessary.
         * 
         * @throws IllegalStateException
         */
        private void validate() throws IllegalStateException {
            if (quoteChar != null && delimiter == quoteChar.charValue()) {
                throw new IllegalStateException("The quoteChar character and the delimiter cannot be the same ('" + quoteChar + "')");
            }
        
            if (escape != null && delimiter == escape.charValue()) {
                throw new IllegalStateException("The escape character and the delimiter cannot be the same ('" + escape + "')");
            }
        
            if (commentStart != null && delimiter == commentStart.charValue()) {
                throw new IllegalStateException("The comment start character and the delimiter cannot be the same ('" + commentStart + 
                        "')");
            }
        
            if (quoteChar != null && quoteChar.equals(commentStart)) {
                throw new IllegalStateException("The comment start character and the quoteChar cannot be the same ('" + commentStart + 
                        "')");
            }
        
            if (escape != null && escape.equals(commentStart)) {
                throw new IllegalStateException("The comment start and the escape character cannot be the same ('" + commentStart + "')");
            }
        
            if (escape == null && quotePolicy == Quote.NONE) {
                throw new IllegalStateException("No quotes mode set but no escape character is set");
            }
        }

        /**
         * Sets the delimiter of the format to the specified character.
         *
         * @param delimiter
         *            the delimiter character
         * @return This builder with the specified character as delimiter
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withDelimiter(final char delimiter) {
            if (isLineBreak(delimiter)) {
                throw new IllegalArgumentException("The delimiter cannot be a line break");
            }
            this.delimiter = delimiter; 
            return this;
        }

        /**
         * Sets the quoteChar of the format to the specified character.
         *
         * @param quoteChar
         *            the quoteChar character
         * @return This builder with the specified character as quoteChar
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withQuoteChar(final char quoteChar) {
            return withQuoteChar(Character.valueOf(quoteChar));
        }

        /**
         * Sets the quoteChar of the format to the specified character.
         *
         * @param quoteChar
         *            the quoteChar character
         * @return This builder with the specified character as quoteChar
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withQuoteChar(final Character quoteChar) {
            if (isLineBreak(quoteChar)) {
                throw new IllegalArgumentException("The quoteChar cannot be a line break");
            }
            this.quoteChar = quoteChar;
            return this;
        }

        /**
         * Sets the comment start marker of the format to the specified character.
         *
         * Note that the comment introducer character is only recognised at the start of a line.
         *
         * @param commentStart
         *            the comment start marker
         * @return This builder with the specified character as the comment start marker
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withCommentStart(final char commentStart) {
            return withCommentStart(Character.valueOf(commentStart));
        }

        /**
         * Sets the comment start marker of the format to the specified character.
         *
         * Note that the comment introducer character is only recognised at the start of a line.
         *
         * @param commentStart
         *            the comment start marker
         * @return This builder with the specified character as the comment start marker
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withCommentStart(final Character commentStart) {
            if (isLineBreak(commentStart)) {
                throw new IllegalArgumentException("The comment start character cannot be a line break");
            }
            this.commentStart = commentStart;
            return this;
        }

        /**
         * Sets the escape character of the format to the specified character.
         *
         * @param escape
         *            the escape character
         * @return This builder with the specified character as the escape character
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withEscape(final char escape) {
            return withEscape(Character.valueOf(escape));
        }

        /**
         * Sets the escape character of the format to the specified character.
         *
         * @param escape
         *            the escape character
         * @return This builder with the specified character as the escape character
         * @throws IllegalArgumentException
         *             thrown if the specified character is a line break
         */
        public CSVFormatBuilder withEscape(final Character escape) {
            if (isLineBreak(escape)) {
                throw new IllegalArgumentException("The escape character cannot be a line break");
            }
            this.escape = escape;
            return this;
        }

        /**
         * Sets the header of the format. The header can either be parsed automatically from the
         * input file with:
         *
         * <pre>
         * CSVFormat format = aformat.withHeader();
         * </pre>
         *
         * or specified manually with:
         *
         * <pre>
         * CSVFormat format = aformat.withHeader(&quot;name&quot;, &quot;email&quot;, &quot;phone&quot;);
         * </pre>
         *
         * @param header
         *            the header, <tt>null</tt> if disabled, empty if parsed automatically, user specified otherwise.
         *
         * @return This builder with the specified header
         */
        public CSVFormatBuilder withHeader(final String... header) {
            this.header = header;
            return this;
        }
        
        /**
         * Sets the trimming behavior of the format.
         *
         * @param ignoreSurroundingSpaces
         *            the trimming behavior, <tt>true</tt> to remove the surrounding spaces, <tt>false</tt> to leave the
         *            spaces as is.
         * @return This builder with the specified trimming behavior.
         */
        public CSVFormatBuilder withIgnoreSurroundingSpaces(final boolean ignoreSurroundingSpaces) {
            this.ignoreSurroundingSpaces = ignoreSurroundingSpaces;
            return this;
        }

        /**
         * Sets the empty line skipping behavior of the format.
         *
         * @param ignoreEmptyLines
         *            the empty line skipping behavior, <tt>true</tt> to ignore the empty lines between the records,
         *            <tt>false</tt> to translate empty lines to empty records.
         * @return This builder with the specified empty line skipping behavior.
         */
        public CSVFormatBuilder withIgnoreEmptyLines(final boolean ignoreEmptyLines) {
            this.ignoreEmptyLines = ignoreEmptyLines;
            return this;
        }

        /**
         * Sets the record separator of the format to the specified character.
         *
         * @param recordSeparator
         *            the record separator to use for output.
         *
         * @return This builder with the the specified output record separator
         */
        public CSVFormatBuilder withRecordSeparator(final char recordSeparator) {
            return withRecordSeparator(String.valueOf(recordSeparator));
        }

        /**
         * Sets the record separator of the format to the specified String.
         *
         * @param recordSeparator
         *            the record separator to use for output.
         *
         * @return This builder with the the specified output record separator
         */
        public CSVFormatBuilder withRecordSeparator(final String recordSeparator) {
            this.recordSeparator = recordSeparator;
            return this;
        }

        /**
         * Sets the output quote policy of the format to the specified value.
         *
         * @param quotePolicy
         *            the quote policy to use for output.
         *
         * @return This builder with the specified quote policy
         */
        public CSVFormatBuilder withQuotePolicy(final Quote quotePolicy) {
            this.quotePolicy = quotePolicy;
            return this;
        }
    }
}
