package org.apache.ivory.entity;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.ivory.IvoryException;
import org.apache.ivory.Tag;
import org.apache.ivory.entity.WorkflowNameBuilder.WorkflowName;
import org.apache.ivory.entity.common.DateValidator;
import org.apache.ivory.entity.store.ConfigurationStore;
import org.apache.ivory.entity.v0.Entity;
import org.apache.ivory.entity.v0.EntityType;
import org.apache.ivory.entity.v0.Frequency;
import org.apache.ivory.entity.v0.feed.Cluster;
import org.apache.ivory.entity.v0.feed.Feed;
import org.apache.ivory.entity.v0.process.Process;

public class EntityUtil {

    private static final DateValidator DateValidator = new DateValidator();
    private static final long MINUTE_IN_MS = 60000L;
    private static final long HOUR_IN_MS = 3600000L;
    private static final long DAY_IN_MS = 86400000L;
    private static final long MONTH_IN_MS = 2592000000L;

    public static Entity getEntity(EntityType type, String entityName) throws IvoryException {
        ConfigurationStore configStore = ConfigurationStore.get();
        Entity entity = configStore.get(type, entityName);
        if (entity == null) {
            throw new EntityNotRegisteredException(entityName + " (" + type + ") not found");
        }
        return entity;        
    }
    
    public static Entity getEntity(String type, String entityName) throws IvoryException {
        EntityType entityType;
        try {
            entityType = EntityType.valueOf(type.toUpperCase());
        } catch (IllegalArgumentException e) {
            throw new IvoryException("Invalid entity type: " + type, e);
        }
        return getEntity(entityType, entityName);
    }

    private static DateFormat getDateFormat() {
        DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm'Z'");
        dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
        return dateFormat;
    }

    public static TimeZone getTimeZone(String tzId) {
        if (tzId == null) {
            throw new IllegalArgumentException("Invalid TimeZone: " + tzId);
        }
        TimeZone tz = TimeZone.getTimeZone(tzId);
        if (!tzId.equals("GMT") && tz.getID().equals("GMT")) {
            throw new IllegalArgumentException("Invalid TimeZone: " + tzId);
        }
        return tz;
    }

    public static Date parseDateUTC(String s) throws IvoryException {
        try {
            return getDateFormat().parse(s);
        } catch (ParseException e) {
            throw new IvoryException(e);
        }
    }

    public static String formatDateUTC(Date d) {
        return (d != null) ? getDateFormat().format(d) : null;
    }

    public static boolean isValidUTCDate(String date) {
        return DateValidator.validate(date);
    }

    public static Date getEndTime(Entity entity, String cluster) throws IvoryException {
        if (entity.getEntityType() == EntityType.PROCESS) {
            return getEndTime((Process) entity, cluster);
        } else {
            return getEndTime((Feed) entity, cluster);
        }
    }

    public static Date getStartTime(Entity entity, String cluster) throws IvoryException {
        if (entity.getEntityType() == EntityType.PROCESS) {
            return getStartTime((Process) entity, cluster);
        } else {
            return getStartTime((Feed) entity, cluster);
        }
    }

    public static Date getEndTime(Process process, String cluster) throws IvoryException {
        return EntityUtil.parseDateUTC(process.getValidity().getEnd());
    }

    public static Date getStartTime(Process process, String cluster) throws IvoryException {
        return EntityUtil.parseDateUTC(process.getValidity().getStart());
    }

    public static Date getEndTime(Feed feed, String cluster) throws IvoryException {
        org.apache.ivory.entity.v0.feed.Cluster clusterDef = FeedHelper.getCluster(feed, cluster);
        return EntityUtil.parseDateUTC(clusterDef.getValidity().getEnd());
    }

    public static Date getStartTime(Feed feed, String cluster) throws IvoryException {
        org.apache.ivory.entity.v0.feed.Cluster clusterDef = FeedHelper.getCluster(feed, cluster);
        return EntityUtil.parseDateUTC(clusterDef.getValidity().getStart());
    }

    public static int getConcurrency(Entity entity) {
        if (entity.getEntityType() == EntityType.PROCESS) {
            return getConcurrency((Process) entity);
        } else {
            return getConcurrency((Feed) entity);
        }
    }

    public static void setStartDate(Entity entity, String cluster, Date startDate) {
        if (entity.getEntityType() == EntityType.PROCESS) {
            setStartDate((Process) entity, cluster, startDate);
        } else {
            setStartDate((Feed) entity, cluster, startDate);
        }
    }

    public static void setEndTime(Entity entity, String cluster, Date endDate) {
        if (entity.getEntityType() == EntityType.PROCESS) {
            setEndTime((Process) entity, cluster, endDate);
        } else {
            setEndTime((Feed) entity, cluster, endDate);
        }
    }

    public static void setConcurrency(Entity entity, int concurrency) {
        if (entity.getEntityType() == EntityType.PROCESS) {
            setConcurrency((Process) entity, concurrency);
        } else {
            setConcurrency((Feed) entity, concurrency);
        }
    }

    public static int getConcurrency(Process process) {
        return process.getConcurrency();
    }

    public static void setStartDate(Process process, String cluster, Date startDate) {
        process.getValidity().setStart(EntityUtil.formatDateUTC(startDate));
    }

    public static void setConcurrency(Process process, int concurrency) {
        process.setConcurrency(concurrency);
    }

    public static void setEndTime(Process process, String cluster, Date endDate) {
        process.getValidity().setEnd(EntityUtil.formatDateUTC(endDate));
    }

    public static int getConcurrency(Feed feed) {
        return 1;
    }

    public static void setStartDate(Feed feed, String cluster, Date startDate) {
        org.apache.ivory.entity.v0.feed.Cluster clusterDef = FeedHelper.getCluster(feed, cluster);
        clusterDef.getValidity().setStart(EntityUtil.formatDateUTC(startDate));
    }

    public static void setEndTime(Feed feed, String cluster, Date endDate) {
        org.apache.ivory.entity.v0.feed.Cluster clusterDef = FeedHelper.getCluster(feed, cluster);
        clusterDef.getValidity().setStart(EntityUtil.formatDateUTC(endDate));
    }

    public static void setConcurrency(Feed feed, int concurrency) {
    }

    public static Date getNextStartTime(Date startTime, Frequency frequency, String timezone, Date now) {
        if (startTime.after(now))
            return startTime;

        Calendar startCal = Calendar.getInstance(EntityUtil.getTimeZone(timezone));
        startCal.setTime(startTime);

        int count = 0;
        switch (frequency.getTimeUnit()) {
            case months:
                count = (int) ((now.getTime() - startTime.getTime()) / MONTH_IN_MS);
                break;
            case days:
                count = (int) ((now.getTime() - startTime.getTime()) / DAY_IN_MS);
                break;
            case hours:
                count = (int) ((now.getTime() - startTime.getTime()) / HOUR_IN_MS);
                break;
            case minutes:
                count = (int) ((now.getTime() - startTime.getTime()) / MINUTE_IN_MS);
                break;
            default:
        }

        if (count > 2) {
            startCal.add(frequency.getTimeUnit().getCalendarUnit(), ((count - 2) / frequency.getFrequency()) * frequency.getFrequency());
        }
        while (startCal.getTime().before(now)) {
            startCal.add(frequency.getTimeUnit().getCalendarUnit(), frequency.getFrequency());
        }
        return startCal.getTime();
    }

    public static int getInstanceSequence(Date startTime, Frequency frequency, String timezone, Date instanceTime) {
        if (startTime.after(instanceTime))
            return -1;

        Calendar startCal = Calendar.getInstance(EntityUtil.getTimeZone(timezone));
        startCal.setTime(startTime);

        int count = 0;
        switch (frequency.getTimeUnit()) {
            case months:
                count = (int) ((instanceTime.getTime() - startTime.getTime()) / MONTH_IN_MS);
                break;
            case days:
                count = (int) ((instanceTime.getTime() - startTime.getTime()) / DAY_IN_MS);
                break;
            case hours:
                count = (int) ((instanceTime.getTime() - startTime.getTime()) / HOUR_IN_MS);
                break;
            case minutes:
                count = (int) ((instanceTime.getTime() - startTime.getTime()) / MINUTE_IN_MS);
                break;
            default:
        }

        if (count > 2) {
            startCal.add(frequency.getTimeUnit().getCalendarUnit(), (count / frequency.getFrequency()) * frequency.getFrequency());
            count = (count / frequency.getFrequency());
        }
        while (startCal.getTime().before(instanceTime)) {
            startCal.add(frequency.getTimeUnit().getCalendarUnit(), frequency.getFrequency());
            count++;
        }
        return count + 1;
    }

    public static String getStagingPath(Entity entity) throws IvoryException {
        try {
            byte[] digest = DigestUtils.md5(entity.toComparableString());
            return "ivory/workflows/" + entity.getEntityType().name().toLowerCase() + "/" + entity.getName() + "/"
                    + new String(Hex.encodeHex(digest));
        } catch (Exception e) {
            throw new IvoryException(e);
        }
    }
    
    public static WorkflowName getWorkflowName(Tag tag, List<String> suffixes,
			Entity entity) {
		WorkflowNameBuilder<Entity> builder = new WorkflowNameBuilder<Entity>(
				entity);
		builder.setTag(tag);
		builder.setSuffixes(suffixes);
		return builder.getWorkflowName();
	}

	public static WorkflowName getWorkflowName(Tag tag, Entity entity) {
		return getWorkflowName(tag, null, entity);
	}

	public static WorkflowName getWorkflowName(Entity entity) {
		return getWorkflowName(null, null, entity);
	}

	public static String getWorkflowNameSuffixes(String workflowName, Entity entity) {
		WorkflowNameBuilder<Entity> builder = new WorkflowNameBuilder<Entity>(
				entity);
		return builder.getWorkflowSuffixes(workflowName);
	}

	public static Tag getWorkflowNameTag(String workflowName, Entity entity) {
		WorkflowNameBuilder<Entity> builder = new WorkflowNameBuilder<Entity>(
				entity);
		return builder.getWorkflowTag(workflowName);
	}
	
	public static String[] getClustersDefined(Entity entity) {
	    switch(entity.getEntityType()) {
	        case CLUSTER:
	            return new String[] { entity.getName() };
	            
	        case FEED:
	            Feed feed = (Feed) entity;
	            List<String> clusters = new ArrayList<String>();
	            for(Cluster cluster:feed.getClusters().getClusters())
	                clusters.add(cluster.getName());
	            return clusters.toArray(new String[clusters.size()]);
	            
	        case PROCESS:
	            Process process = (Process) entity;
	            return new String[] { process.getCluster().getName() };
	    }  
	    throw new IllegalArgumentException("Unhandled entity type: " + entity.getEntityType());
	}
}
